use std::{cell::RefCell, collections::HashMap, rc::Rc};

use super::SizePolicy;

#[derive(PartialEq, Eq, PartialOrd, Ord, Hash, Debug, Clone, Copy)]
pub struct LayoutCacheKey(usize);
static NEXT_CACHE_KEY: std::sync::atomic::AtomicUsize = std::sync::atomic::AtomicUsize::new(1);
impl LayoutCacheKey {
    pub fn generate() -> Self {
        Self(NEXT_CACHE_KEY.fetch_add(1, std::sync::atomic::Ordering::SeqCst))
    }
}

#[derive(Clone, Debug)]
pub struct NodeState {
    pub(super) needs_update: bool,
    pub(super) needs_render: bool,
    pub(super) net_policy: (SizePolicy, SizePolicy),
    pub(super) layer: Layer,
    pub(super) area: Option<kahlo::math::PixelBox>,
    pub(super) children: Vec<LayoutCacheKey>,
}

#[derive(Debug, Default)]
pub struct LayoutCache {
    pub(super) update_queue: RefCell<Vec<LayoutCacheKey>>,
    pub(super) render_queue: RefCell<Vec<LayoutCacheKey>>,
    states: RefCell<HashMap<LayoutCacheKey, NodeState>>,
    parents: RefCell<HashMap<LayoutCacheKey, LayoutCacheKey>>,
}

impl LayoutCache {
    pub fn new() -> Rc<LayoutCache> {
        Rc::new(Self {
            update_queue: Default::default(),
            render_queue: Default::default(),
            states: Default::default(),
            parents: Default::default(),
        })
    }

    pub fn propagate_relayouts(&self) {
        let mut updq = self.update_queue.borrow_mut();
        let mut states = self.states.borrow_mut();
        let parents = self.parents.borrow();
        while let Some(next) = updq.pop() {
            if let Some(v) = states.get_mut(&next) {
                if !v.needs_update {
                    if let Some(p) = parents.get(&next) {
                        updq.push(*p)
                    }
                    v.needs_update = true;
                }
            }
        }
    }

    pub fn propagate_rerenders(&self) {
        let mut renq = self.render_queue.borrow_mut();
        let mut states = self.states.borrow_mut();

        while let Some(next) = renq.pop() {
            if let Some(v) = states.get_mut(&next) {
                v.needs_render = true;
                renq.extend(v.children.iter());
            }
        }
    }

    pub fn with_state<R, F: FnOnce(&mut NodeState) -> R>(
        &self,
        ckey: LayoutCacheKey,
        f: F,
    ) -> Option<R> {
        self.states.borrow_mut().get_mut(&ckey).map(f)
    }

    pub fn has_state_for(&self, ckey: LayoutCacheKey) -> bool {
        self.states.borrow().contains_key(&ckey)
    }

    pub fn store(&self, ckey: LayoutCacheKey, parent: Option<LayoutCacheKey>, ns: NodeState) {
        self.states.borrow_mut().insert(ckey, ns);
        if let Some(pck) = parent {
            self.parents.borrow_mut().insert(ckey, pck);
        }
    }
}

#[derive(Clone, Copy, PartialEq, PartialOrd, Debug, Default)]
pub struct Layer(u16, u16);

impl Layer {
    pub fn nest(&self) -> Self {
        Layer(self.0, self.1 + 1)
    }

    pub fn float(&self) -> Self {
        Layer(self.0 + 1, 0)
    }
}
