use kahlo::{
    math::{PixelBox, PixelPoint, PixelSize},
    prelude::*,
};

pub struct TextLine<'l> {
    line: &'l str,
    font: &'l fontdue::Font,
    size: f32,
}

impl<'l> TextLine<'l> {
    pub(crate) fn new(font: &'l fontdue::Font, line: &'l str, size: f32) -> Self {
        Self { line, font, size }
    }

    fn generate_baseline_offsets(&self) -> impl '_ + Iterator<Item = (u32, u32, char)> {
        let mut running_offset = 0;
        let mut last_char = None;

        self.line.chars().map(move |ch| {
            let metrics = self.font.metrics(ch, self.size);

            let glyph_width = if let Some(last) = last_char.take() {
                metrics.advance_width + self.font.horizontal_kern(last, ch, self.size).unwrap_or(0.)
            } else {
                metrics.advance_width
            };

            last_char = Some(ch);

            let expose = (running_offset, glyph_width.round() as u32, ch);
            running_offset += glyph_width.round() as u32;
            expose
        })
    }

    pub fn calculate_size(&self) -> PixelSize {
        // pass: determine required size
        let total_width = {
            let ofs = self.generate_baseline_offsets().last().unwrap();
            ofs.0 + ofs.1
        };
        let line_metrics = self.font.horizontal_line_metrics(self.size).unwrap();
        let line_height = line_metrics.new_line_size.ceil() as i32;

        PixelSize::new(total_width as i32, line_height)
    }

    pub fn render_line(&self) -> kahlo::Alphamap {
        // pass 1: determine required size
        let alphamap_size = self.calculate_size();

        let line_metrics = self.font.horizontal_line_metrics(self.size).unwrap();
        let baseline = line_metrics.ascent.ceil() as usize;
        let mut alphamap =
            kahlo::Alphamap::new(alphamap_size.width as usize, alphamap_size.height as usize);

        // pass: generate alphamap
        for (offset, _width, ch) in self.generate_baseline_offsets() {
            let (metrics, raster) = self.font.rasterize(ch, self.size);
            let character_alphamap = kahlo::BitmapRef::<kahlo::formats::A8>::new(
                raster.as_slice(),
                metrics.width,
                metrics.height,
            );

            alphamap.copy_from(
                &character_alphamap,
                PixelBox::from_size(character_alphamap.size()),
                PixelPoint::new(
                    metrics.xmin + offset as i32,
                    baseline as i32 - (metrics.height as i32 + metrics.ymin),
                ),
            );
        }

        alphamap
    }
}

pub struct TextBuffer {
    lines: Vec<String>,
}

impl TextBuffer {}
