use std::ops::DerefMut;

use kahlo::prelude::*;

use crate::{
    component::Component,
    input::InputState,
    layout::{LayoutNode, LayoutNodeAccess, LeafLayoutNode, SizePolicy},
    theme::Theme,
    ui::UIHandle,
    window::RenderFormat,
};

mod label;
pub use label::Label;
mod frame;
pub use frame::Frame;
mod group;
pub use group::PlainGroup;
mod button;
pub use button::Button;

pub trait Widget<C: Component> {
    fn poll(&mut self, uih: &mut UIHandle, input_state: Option<&InputState>) -> Vec<C::Msg>;

    fn layout_node(&self) -> LayoutNodeAccess;
    fn layout_node_mut(&mut self) -> &mut LayoutNode;
    fn render(&self, theme: &Theme, target: &mut kahlo::BitmapMut<RenderFormat>);
}

pub struct Spacer<C: Component> {
    lnode: LeafLayoutNode,
    _ghost: std::marker::PhantomData<C>,
}

impl<C: Component> Spacer<C> {
    pub fn new(uih: &UIHandle) -> Self {
        let mut lnode = LeafLayoutNode::new(uih.new_layout_node());
        lnode
            .set_width_policy(SizePolicy::expanding(1))
            .set_height_policy(SizePolicy::expanding(1));
        Self {
            lnode,
            _ghost: Default::default(),
        }
    }
}

impl<C: Component> Widget<C> for Spacer<C> {
    fn poll(
        &mut self,
        _uih: &mut UIHandle,
        _input_state: Option<&InputState>,
    ) -> Vec<<C as Component>::Msg> {
        vec![]
    }

    fn layout_node(&self) -> LayoutNodeAccess {
        LayoutNodeAccess::new(&self.lnode)
    }
    fn layout_node_mut(&mut self) -> &mut LayoutNode {
        self.lnode.deref_mut()
    }

    fn render(&self, theme: &Theme, target: &mut kahlo::BitmapMut<RenderFormat>) {
        let region = self.layout_node().render_area().unwrap();
        if self.lnode.render_check() {
            target.fill_region(region, theme.background);
        }
    }
}
